/* Potential flow functions for vortex2d */

#include "const.h"
#include "structs.h"
#include <math.h>
#include <stdio.h>

extern double kinematic_viscosity;
extern int num_e, num_p, num_o;
extern ELEMENT e[];
extern PARTICLE p[];
extern OBJECT o[];

VEC ne[MAXELEMENTS];	/* temporary array of new element locations */
VEC tot_inf, inf;	/* global declarations of oft-used vectors */

int move_everything(double);
void total_influence(VEC);
void element_influence(ELEMENT, VEC);
void decay_vortex_blobs(double);

/*
 * move all objects as required
 */
int move_everything(double dt) {

   register int i,j;
   register double xinf,yinf;

   /* move all particles */
   for (i=0; i<num_p; i++) {
      /* calculate total influence, put it in tot_inf */
      /* total_influence(p[i].loc); */
      xinf = 0.0;
      yinf = 0.0;
      for (j=0; j<num_e; j++) {
         /* influence of element j on loc of particle i */
         (void) element_influence(e[j],p[i].loc);
         xinf += inf.x[0];
         yinf += inf.x[1];
      }
      p[i].loc.x[0] += xinf*dt;
      p[i].loc.x[1] += yinf*dt;
   }

   /* move all objects */
   for (i=0; i<num_o; i++) {
      /* stuff */
   }

   /* move all elements */
   /* first, find influences using old locations */
   for (i=0; i<num_e; i++) {
      /* calculate total influence on element i, put it in ne */
      xinf = 0.0;
      yinf = 0.0;
      for (j=0; j<num_e; j++) {
         if (i != j) {
            /* influence of element j on loc of element i */
            (void) element_influence(e[j],e[i].loc);
            xinf += inf.x[0];
            yinf += inf.x[1];
         }
      }
      ne[i].x[0] = xinf*dt;
      ne[i].x[1] = yinf*dt;
   }
   /* then, apply all changes to elements' locations */
   for (i=0; i<num_e; i++) {
      e[i].loc.x[0] += ne[i].x[0];
      e[i].loc.x[1] += ne[i].x[1];
   }
   /* lastly, update all vortex blobs' char. radius */
   decay_vortex_blobs(dt);

   return(1);
}


/*
 * compute the influence of all elements at a point
 */
void total_influence(VEC point) {

   int i;

   /* clear out what's stored away */
   tot_inf.x[0] = 0.0;
   tot_inf.x[1] = 0.0;

   for (i=0; i<num_e; i++) {
      (void) element_influence(e[i],point);
      tot_inf.x[0] += inf.x[0];
      tot_inf.x[1] += inf.x[1];
   }
}


/*
 * compute the influence of an element on a point
 */
void element_influence(ELEMENT element, VEC point) {

   double dx, dy, dist, dist_squared, dist_cubed, scale;
   inf.x[0] = 0.0;
   inf.x[1] = 0.0;

   switch (element.type) {
      case 'f':		/* freestream */
         inf.x[0] = cos(element.val) * element.str;
         inf.x[1] = sin(element.val) * element.str;
         break;

      /* need these values for all other elements */

      case 's':		/* source/sink */
         dx = point.x[0] - element.loc.x[0];
         dy = point.x[1] - element.loc.x[1];
         dist_squared = dx*dx + dy*dy;
         if (dist_squared < SM_THRESH) {
            inf.x[0] = 0.0;
            inf.x[1] = 0.0;
            break;
         }
         dist_cubed = sqrt(dist_squared)*dist_squared;
         scale = TWOPI * element.str / dist_squared;
         inf.x[0] = dx * scale;
         inf.x[1] = dy * scale;
         break;

      case 'v':		/* standard vortex */
         dx = point.x[0] - element.loc.x[0];
         dy = point.x[1] - element.loc.x[1];
         dist_squared = dx*dx + dy*dy;
         if (dist_squared < SM_THRESH) {
            inf.x[0] = 0.0;
            inf.x[1] = 0.0;
            break;
         }
         dist_cubed = sqrt(dist_squared)*dist_squared;
         scale = TWOPI * element.str / dist_cubed;
         inf.x[0] = -dy * scale;
         inf.x[1] = dx * scale;
         break;

      case 'b':		/* vortex blob */
         /*
          * In a complex vortex, I try to simulate the center
          * of the vortex as moving slower, and not faster, the
          * closer to the middle you get. element.val is the
          * distance from the vortex where the magnitude of the
          * influence is greatest. Here, element.str is the 
          * value of the magnitude of the influence at that
          * distance. The critical distance is the distance at
          * which magnitude of the influence is greatest.
          */
         /*
          * Addendum: a vortex blob is defined by a strength and
          * a characteristic distance^2. The characteristic distance
          * is the distance at which the influence is greatest.
          * BUT, the maximum magnitude of the influence varies 
          * with the characteristic distance. This is done in
          * order to keep the physical properties of a constant-
          * strength vortex (constant momentum stuff)
          */
         dx = point.x[0] - element.loc.x[0];
         dy = point.x[1] - element.loc.x[1];
         /* this is the distance normalized to critical distance */
         /* dist_squared = (dx*dx + dy*dy) / (element.val*element.val); */
         /* the new formulation needs the raw distance */
         dist_squared = (dx*dx + dy*dy);
         dist = sqrt(dist_squared);
         /* This is my old formula */
         /* scale = TWOPI * element.str / (dist*(dist_squared - 2) + 2); */
         /* This is the Gaussian distribution: */
         scale = element.str * exp(-dist_squared/element.val) / (PI * element.val) / dist;
         inf.x[0] = -dy * scale;
         inf.x[1] = dx * scale;
         break;

      default:		/* do nothing */
         inf.x[0] = 0.0;
         inf.x[1] = 0.0;
         break;
   }	/* end switch on element.type */

   /* fprintf(stderr,"inf of element at %lf,%lf on point %lf,%lf is %lf,%lf\n",
         element.loc.x[0],element.loc.x[1],point.x[0],point.x[1],inf.x[0],inf.x[1]); */
}

/*
 * A vortex blob in a real flow decays according to a simple
 * formula. When this function is called, the vortex blobs
 * in the simulation decay.
 */
void decay_vortex_blobs(double delta_t)
{
   int i;

   for (i=0; i<num_e; i++) if (e[i].type == 'b') {
      e[i].val = e[i].val + 4*delta_t*kinematic_viscosity;
   }
}

