/* ------------------------------------------------------------------
 *
 * vortex2d.c
 *
 * copyright 1998 Mark J. Stock
 *
 * v0.0 1998-05-28	initial version, free vortex blobs
 * v0.1 1998-06-12	generators work, predictor-corrector scheme
 *
 * This program will eventually simulate a two-dimensional potential
 * flow with a variety of flow elements, including vortexes, sources,
 * a freestream, and fixed and moving objects.
 *
 * It will also support export of particles and objects to a Radiance-
 * readable format, for visualization. In addition, a simple textual
 * output option will be made available.
 *
 * ------------------------------------------------------------------ */


#include "const.h"
#include "structs.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/*
extern int set_up_space(double,double);
extern int set_up_stuff(int);
extern void write_error(char,int);
extern int write_text(int,int,double);
extern int move_everything(double);
*/

SPACE   sp;             /* the dimensions of the area in question */

/* the array of fluid dynamic singularities */
ELEMENT e[MAX_ELEMENTS];
int     num_e;          /* number of elements */

/* the array of vortex generators */
ELEMENT_GENERATOR eg[MAX_ELEMENT_GENERATORS];
int     num_eg;          /* number of generators */

/* the array of particles to animate */
PARTICLE_OLD p[MAX_PARTICLES];
int     num_p;          /* number of particles */

/* the array of particle generators */
PARTICLE_GENERATOR pg[MAX_PARTICLE_GENERATORS];
int     num_pg;

OBJECT  o[MAX_OBJECTS];  /* the array of solid_objects */
int     num_o;          /* number of objects */

double kinematic_viscosity;

double start_time;      /* simulation start time */
double end_time;        /* simulation end time */
double curr_time;       /* current simulation time */


main(int argc, char *argv[]) {

   /* initialize variables */
   int i;
   int add_p = 0;			/* number of particles to add */
   int add_v = 0;			/* number of vortexes to add */
   double xsize = 6.0;
   double ysize = 4.0;
   double compute_dt = 0.01;		/* compute every n sec */
   double write_dt = 0.25;		/* write every n sec */
   double last_write_time = -1.0;	/* last time output was written */
   int framenum = 0;			/* current frame number */
   int output_type = 1;			/* 0 is none, 1 is text, 2 is radiance, 3 is PGM */
   int output_size = 72;		/* default output size */
   char progname[80];			/* name of executable */
   int seed = 0;			/* seed for rand # generator */
   int temp1, temp2;

   /* some physical constants */
   kinematic_viscosity = 1.46e-5;
   start_time = 0.0;
   end_time = 10.0;
   curr_time = start_time;

   /*
    * Parse the command-line options
    */
   (void) strcpy(progname,argv[0]);
   for (i=1; i<argc; i++) {
      if (strncmp(argv[i], "-help", 2) == 0)
         (void) usage(progname,0);
      else if (strncmp(argv[i], "-end", 2) == 0)
         end_time = atof(argv[++i]);	/* reset simulation end time */
      else if (strncmp(argv[i], "-cdt", 2) == 0)
         compute_dt = atof(argv[++i]);	/* reset compute delta time */
      else if (strncmp(argv[i], "-odt", 2) == 0)
         write_dt = atof(argv[++i]);	/* reset output delta time */
      else if (strncmp(argv[i], "-size", 3) == 0) {
         xsize = atof(argv[++i]);	/* modify solution area */
         ysize = atof(argv[++i]); }
      else if (strncmp(argv[i], "-radiance", 2) == 0)
         output_type = 2;		/* write output to Radiance format */
      else if (strncmp(argv[i], "-text", 2) == 0)
         output_type = 1;		/* write text output to screen */
      else if (strncmp(argv[i], "-pgm", 2) == 0)
         output_type = 3;		/* write text output to PGM image file */
      else if (strncmp(argv[i], "-width", 2) == 0)
         output_size = atoi(argv[++i]);	/* reset image or text output width */
      else if (strncmp(argv[i], "-np", 3) == 0)
         add_p = atoi(argv[++i]);	/* add a few more particles */
      else if (strncmp(argv[i], "-nv", 3) == 0)
         add_v = atoi(argv[++i]);	/* add a few more vortexes */
      else if (strncmp(argv[i], "-seed", 3) == 0)
         seed = atoi(argv[++i]);	/* reset the rnd # seed */
      else if (strncmp(argv[i], "-kv", 2) == 0)
         kinematic_viscosity = atof(argv[++i]);	/* reset kinematic viscosity */
      else
         (void) usage(progname,0);
   }

   /*
    * Set up the problem space
    */
   for (i=0; i<seed; i++) (void) rand();
   if (set_up_space(xsize,ysize)) write_error("Error setting up compute area.",1);
   if (set_up_stuff(4)) write_error("Error setting up stuff.",1);
   if (add_v > 0) add_random_vortexes(add_v,5.0);
   if (add_p > 0) add_random_particles(add_p,output_size);
   /* write_element_list(); */
   /* write_particle_list(); */
   /* write_emitter_list(); */

   /*
    * Begin the loop over time
    */
   while (curr_time < end_time + SM_THRESH) {

      /*
       * Check to see if output should be written
       */
      if (curr_time > last_write_time + write_dt - SM_THRESH) {
         (void) write_output(output_type,output_size,framenum);
         last_write_time = curr_time;
         if (output_type > 1) fprintf(stderr,"%d\n",framenum);
         framenum++;
         if (output_type > 1) fflush(stderr);
         if (last_write_time + write_dt > end_time + SM_THRESH) break;
      }


      /*
       * Allow generators to create new elements
       */
      if (num_eg > 0) {
         temp1 = generate_elements(compute_dt);
         /* fprintf(stderr,"%d elements generated\n",temp1); */
      }
      if (num_pg > 0) {
         temp2 = generate_particles(compute_dt);
         /* fprintf(stderr,"%d particles generated\n",temp2); */
      }


      /*
       * If objects exist, solve for their element strengths
       */
      if (num_o > 0) {
         /* (void) solve_object_strengths(accuracy_param); */
      }


      /*
       * Move all things in the flowfield
       */
      (void) move_everything(compute_dt);


      /*
       * Update simulation time and go back
       */
      curr_time += compute_dt;
      if (output_type > 1) {
         fprintf(stderr,".");
         fflush(stderr);
      }
      /* (void) write_stats(); */

   }	/* end while keep_going */

   /*
    * Write output statistics and quit
    */
   (void) write_stats();
   (void) write_finish();
}


int write_output(int output_type,int output_size,int framenum) {

   char filename[80];			/* a filename storage thing */

   switch (output_type) {
      case 0:		/* no ouput at all */
         break;
      case 1:		/* write text to stdout */
         fprintf(stderr,"Press return for next frame.");
         (void) getchar();
         (void) write_text(2,output_size,curr_time);
         break;
      case 2:		/* write a Radiance file */
         if (framenum < 10) sprintf(filename,"frame0%d.rad",framenum);
         else sprintf(filename,"frame%d.rad",framenum);
         (void) write_radiance(filename,curr_time);
         break;
      case 3:		/* write a PGM file */
         /* if (framenum < 10) sprintf(filename,"frame0%d.pgm",framenum);
         else sprintf(filename,"frame%d.pgm",framenum); */
         sprintf(filename,"frame%03d.pgm",framenum);
         (void) write_pgm(filename,output_size,curr_time);
         break;
     default:		/* no ouput at all */
         break;
   }	/* end switch on output_type */

}
