/* Potential flow functions for vortex2d */

#include "const.h"
#include "structs.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>

extern SPACE sp;
extern int num_e, num_eg, num_p, num_pg, num_o;
extern ELEMENT e[];
extern ELEMENT_GENERATOR eg[];
extern particle_pointer particle_head;
extern PARTICLE_GENERATOR pg[];
extern OBJECT o[];

extern double kinematic_viscosity;
extern double curr_time;

VEC tot_inf, inf;	/* global declarations of oft-used vectors */

int move_everything(double);
void total_influence(VEC);
void element_influence(ELEMENT, VEC);

/*
 * move all objects as required
 */
int move_everything(double dt) {

   register int i,j;
   register double xinf,yinf;
   ELEMENT ne[MAX_ELEMENTS];	/* temp array of new elements */
   VEC oi[MAX_ELEMENTS];		/* old influences, to average w/ new inf */
   register VEC tl;		/* temporary location */
   register double oldx, oldy;
   particle_pointer particle = particle_head;

   /* move all elements - using a predictor-corrector method */
   /* first, create elements at new locations using Euler */
   /* Note, don't move the freestreams, loc is their velocity! */
   for (i=0; i<num_e; i++) if (e[i].index < 0) {
      /* calculate total influence on element i, put it in ne */
      xinf = 0.0;
      yinf = 0.0;
      for (j=0; j<num_e; j++) {
         if (i != j) {
            /* influence of element j on loc of element i */
            (void) element_influence(e[j],e[i].loc);
            xinf += inf.x[0];
            yinf += inf.x[1];
         }
      }
      /* this is the Euler method...one big time step */
      oi[i].x[0] = xinf*dt;
      oi[i].x[1] = yinf*dt;
   }
   /* yeah, create the elements at the new time */
   for (i=0; i<num_e; i++) if (e[i].index < 0) {
      ne[i].type = e[i].type;
      ne[i].loc.x[0] = e[i].loc.x[0] + oi[i].x[0];
      ne[i].loc.x[1] = e[i].loc.x[1] + oi[i].x[1];
      ne[i].str = e[i].str;
      ne[i].val = e[i].val;
      /* vortex blobs decay according to a simple formula: */
      if (ne[i].type == 'b') ne[i].val += 4*dt*kinematic_viscosity;
   }
   /* now, find the influence on each element using the new locations,
    * and average it with the old influences to find the corrected
    * influence; then move ne[] to the actual new locations */
   for (i=0; i<num_e; i++) if (e[i].index < 0) {
      xinf = 0.0;
      yinf = 0.0;
      for (j=0; j<num_e; j++) {
         if (i != j) {
            /* influence of element j on loc of element i */
            (void) element_influence(ne[j],ne[i].loc);
            xinf += inf.x[0];
            yinf += inf.x[1];
         }
      }
      oi[i].x[0] = (oi[i].x[0] + xinf*dt) / 2.0;
      oi[i].x[1] = (oi[i].x[1] + yinf*dt) / 2.0;
   }
   /* resetting the new elements' locations from their old locations
    * using the avarage of the two influence calculations */
   for (i=0; i<num_e; i++) if (e[i].index < 0) {
      ne[i].loc.x[0] = e[i].loc.x[0] + oi[i].x[0];
      ne[i].loc.x[1] = e[i].loc.x[1] + oi[i].x[1];
   }
   /* now, e[] contains the old elements' characteristics, and
    * ne[] contains the characteristics at the end of the time step */

   /* move all particles */
   /* again, use a predictor-corrector method, and now that we have
    * good values for the elements at both ends of the time step,
    * this should be easy */
   while (particle->next_particle != NULL) {
      if (particle->off_time > curr_time) {
         /* calculate total influence from all original elements */
         xinf = 0.0;
         yinf = 0.0;
         for (j=0; j<num_e; j++) {
            /* influence of element j on loc of particle i */
            (void) element_influence(e[j],particle->loc);
            xinf += inf.x[0];
            yinf += inf.x[1];
         }
         /* create a temporary particle at the first guess location */
         oldx = xinf*dt;
         oldy = yinf*dt;
         tl.x[0] = particle->loc.x[0] + oldx;
         tl.x[1] = particle->loc.x[1] + oldy;
         /* calculate total influence using new element locations
            and first guess particle location */
         xinf = 0.0;
         yinf = 0.0;
         for (j=0; j<num_e; j++) {
            (void) element_influence(ne[j],tl);
            xinf += inf.x[0];
            yinf += inf.x[1];
         }
         /* average the two movements to get actual movement */
         particle->loc.x[0] += (xinf*dt + oldx) / 2.0;
         particle->loc.x[1] += (yinf*dt + oldy) / 2.0;
      }
      particle = particle->next_particle;
   }

   /* move all objects */
   for (i=0; i<num_o; i++) {
      /* stuff */
   }

   /* then, apply all changes to elements' locations */
   /* e[] becomes the current element characteristics */
   for (i=0; i<num_e; i++) if (e[i].index < 0) {
      e[i].loc.x[0] = ne[i].loc.x[0];
      e[i].loc.x[1] = ne[i].loc.x[1];
      e[i].val = ne[i].val;
   }

   return(1);
}


/*
 * compute the influence of all elements at a point
 */
void total_influence(VEC point) {

   int k;

   /* clear out what's stored away */
   tot_inf.x[0] = 0.0;
   tot_inf.x[1] = 0.0;

   for (k=0; k<num_e; k++) {
      (void) element_influence(e[k],point);
      /* fprintf(stderr,"influence of element %d is %lf %lf\n",k,inf.x[0],inf.x[1]); */
      tot_inf.x[0] += inf.x[0];
      tot_inf.x[1] += inf.x[1];
   }
}


/*
 * compute the influence of an element on a point
 */
void element_influence(ELEMENT element, VEC point) {

   double dx, dy, dist, dist_squared, dist_cubed, scale;
   inf.x[0] = 0.0;
   inf.x[1] = 0.0;

   switch (element.type) {
      case 'f':		/* freestream */
         inf.x[0] = element.loc.x[0];
         inf.x[1] = element.loc.x[1];
         break;

      /* need these values for all other elements */

      case 's':		/* source/sink */
         dx = point.x[0] - element.loc.x[0];
         dy = point.x[1] - element.loc.x[1];
         dist_squared = dx*dx + dy*dy;
         if (dist_squared < SM_THRESH) {
            inf.x[0] = 0.0;
            inf.x[1] = 0.0;
            break;
         }
         dist_cubed = sqrt(dist_squared)*dist_squared;
         scale = TWOPI * element.str / dist_squared;
         inf.x[0] = dx * scale;
         inf.x[1] = dy * scale;
         break;

      case 'v':		/* standard vortex */
         dx = point.x[0] - element.loc.x[0];
         dy = point.x[1] - element.loc.x[1];
         dist_squared = dx*dx + dy*dy;
         if (dist_squared < SM_THRESH) {
            inf.x[0] = 0.0;
            inf.x[1] = 0.0;
            break;
         }
         dist_cubed = sqrt(dist_squared)*dist_squared;
         scale = TWOPI * element.str / dist_cubed;
         inf.x[0] = -dy * scale;
         inf.x[1] = dx * scale;
         break;

      case 'b':		/* vortex blob */
         /*
          * In a complex vortex, I try to simulate the center
          * of the vortex as moving slower, and not faster, the
          * closer to the middle you get. element.val is the
          * distance from the vortex where the magnitude of the
          * influence is greatest. Here, element.str is the 
          * value of the magnitude of the influence at that
          * distance. The critical distance is the distance at
          * which magnitude of the influence is greatest.
          */
         /*
          * Addendum: a vortex blob is defined by a strength and
          * a characteristic distance^2. The characteristic distance
          * is the distance at which the influence is greatest.
          * BUT, the maximum magnitude of the influence varies 
          * with the characteristic distance. This is done in
          * order to keep the physical properties of a constant-
          * strength vortex (constant momentum stuff)
          */
         dx = point.x[0] - element.loc.x[0];
         dy = point.x[1] - element.loc.x[1];
         /* this is the distance normalized to critical distance */
         /* dist_squared = (dx*dx + dy*dy) / (element.val*element.val); */
         /* the new formulation needs the raw distance */
         dist_squared = (dx*dx + dy*dy);
         dist = sqrt(dist_squared);
         /* This is my old formula */
         /* scale = TWOPI * element.str / (dist*(dist_squared - 2) + 2); */
         /* This is the Gaussian distribution: */
         scale = element.str * exp(-dist_squared/element.val) / (PI * element.val) / dist;
         inf.x[0] = -dy * scale;
         inf.x[1] = dx * scale;
         break;

      default:		/* do nothing */
         inf.x[0] = 0.0;
         inf.x[1] = 0.0;
         break;
   }	/* end switch on element.type */

   /* fprintf(stderr,"inf of element at %lf,%lf on point %lf,%lf is %lf,%lf\n",
         element.loc.x[0],element.loc.x[1],point.x[0],point.x[1],inf.x[0],inf.x[1]); */
}


/*
 * allow each element generator to whip out a few elements
 */
int generate_elements(double dt) {

   int i, j, k;
   int num_new_e = 0;
   int total_new_e = 0;
   double moddt;

   for (i=0; i<num_eg; i++) {
      num_new_e = eg[i].rate * dt;
      if (((double)rand())/RAND_MAX < (eg[i].rate*dt - num_new_e)) num_new_e++;
      /* fprintf(stderr,"rate= %lf, dt= %lf, num_new_e = %d\n",eg[i].rate,dt,num_new_e); */
      if (num_e+num_new_e > MAX_ELEMENTS) return(-1);
      (void) total_influence(eg[i].loc);
      for (j=num_e; j<num_e+num_new_e; j++) {
         moddt = dt*dt * (double)(j-num_e)/(double)num_new_e;
         e[j].index = -1;
         e[j].type = eg[i].model.type;
         e[j].loc.x[0] = eg[i].loc.x[0] + tot_inf.x[0]*moddt;
         e[j].loc.x[1] = eg[i].loc.x[1] + tot_inf.x[1]*moddt;
         e[j].str = eg[i].model.str / eg[i].rate;
         e[j].val = eg[i].model.val;
         /* fprintf(stderr,"made element, loc= %lf %lf, str= %lf\n",e[j].loc.x[0],e[j].loc.x[1],e[j].str); */
      }
      num_e += num_new_e;
      total_new_e += num_new_e;
   }

   return(total_new_e);
}


/*
 * allow each particle generator to whip out a few particles
 */
int generate_particles(double dt) {

   int i,j;
   int num_new_p = 0;
   int total_new_p = 0;
   double angle, rad, moddt;
   particle_pointer old_particle_head;
   particle_pointer particle;

   for (i=0; i<num_pg; i++) {
      num_new_p = pg[i].rate * dt;
      if (rand()/(RAND_MAX+1.0) < (pg[i].rate*dt - num_new_p)) num_new_p++;
      /* fprintf(stderr,"rate= %lf, dt= %lf, num_new_p = %d\n",pg[i].rate,dt,num_new_p); */
      if (num_p+num_new_p > MAX_PARTICLES) num_new_p = MAX_PARTICLES - num_p;
      /* find local fluid velocity */
      (void) total_influence(pg[i].loc);
      /* fprintf(stderr,"total_influence returns %lf %lf at %lf %lf\n",tot_inf.x[0],tot_inf.x[1],pg[i].loc.x[0],pg[i].loc.x[1]); */
      for (j=0; j<num_new_p; j++) {
         angle = (TWOPI*rand())/RAND_MAX;
         rad = (pg[i].rad*rand())/RAND_MAX;
         moddt = dt*dt * j/(double)num_new_p;
         /* fprintf(stderr,"setup, angle= %lf, rad= %lf, moddt= %lf\n",angle,rad,moddt); */
         particle = (PARTICLE *)malloc(sizeof(PARTICLE));
         particle->loc.x[0] = pg[i].loc.x[0] + cos(angle)*rad + tot_inf.x[0]*moddt;
         particle->loc.x[1] = pg[i].loc.x[1] + sin(angle)*rad + tot_inf.x[1]*moddt;
         particle->off_time = (pg[i].age_high-pg[i].age_low)*rand()/(RAND_MAX+1.0);
         particle->off_time += curr_time + pg[i].age_low;
         /* add this particle to the head of the list! */
         old_particle_head = particle_head;
         particle_head = particle;
         particle->next_particle = old_particle_head;
         /* fprintf(stderr,"made particle, loc= %lf %lf, age= %lf\n",particle->loc.x[0],particle->loc.x[1],particle->off_time); */
         num_p++;
      }
      total_new_p += num_new_p;
   }

   return(total_new_p);
}


/*
 * Remove all particles scheduled to die before kill_time
 */
int kill_particles(double kill_time) {

   int num_killed = 0;
   particle_pointer particle = particle_head;
   particle_pointer last_particle = NULL;

   /* if no particles at all, don't kill any! */
   if (particle_head == NULL) return(0);

   while (particle->next_particle != NULL) {
      if (particle->off_time < kill_time) {
         /* kill it */
         if (last_particle == NULL) {
            /* means the first particle needs to be killed */
            particle_head = particle->next_particle;
            free(particle);
            particle = particle_head;
         } else {
            /* skip over the damn particle and free its memory */
            last_particle->next_particle = particle->next_particle;
            free(particle);
            particle = last_particle->next_particle;
         }
         num_killed++;
      } else {
         /* go on to the next one */
         last_particle = particle;
         particle = particle->next_particle;
      }
   }

   num_p = num_p - num_killed;
   return(num_killed);
}

/*
 * Remove all particles in the linked list whose location
 * is within the space defined...area kill
 */
int trim_particles(SPACE area) {

   int num_trimmed = 0;
   particle_pointer particle = particle_head;
   particle_pointer last_particle = NULL;

   /* if no particles at all, don't trim any! */
   if (particle_head == NULL) return(0);

   while (particle->next_particle != NULL) {
      if (particle->loc.x[0] < area.o.x[0] || particle->loc.x[0] > (area.o.x[0]+area.s.x[0])
           || particle->loc.x[1] < area.o.x[1] || particle->loc.x[1] > (area.o.x[1]+area.s.x[1])) {
         /* trim it */
         /* fprintf(stderr,"particle trimmed b/c x= %lf, y= %lf\n",particle->loc.x[0],particle->loc.x[1]); */
         if (last_particle == NULL) {
            /* means the first particle needs to be trimmed */
            particle_head = particle->next_particle;
            free(particle);
            particle = particle_head;
         } else {
            /* skip over the damn particle and free its memory */
            last_particle->next_particle = particle->next_particle;
            free(particle);
            particle = last_particle->next_particle;
         }
         num_trimmed++;
      } else {
         /* go on to the next one */
         last_particle = particle;
         particle = particle->next_particle;
      }
   }

   num_p = num_p - num_trimmed;
   return(num_trimmed);
}


/*
 * Make all elements outside of element_trim become
 * dummy elements, they are not computed
 */
int trim_elements(SPACE area) {

   int i;
   int num_trimmed = 0;

   for (i=0; i<num_e; i++) {
      if (e[i].loc.x[0] < area.o.x[0] || e[i].loc.x[0] > (area.o.x[0]+area.s.x[0])
           || e[i].loc.x[1] < area.o.x[1] || e[i].loc.x[1] > (area.o.x[1]+area.s.x[1])) {
         /* trim it */
         /* fprintf(stderr,"element trimmed b/c x= %lf, y= %lf\n",e[i].loc.x[0],e[i].loc.x[1]); */
         e[i].type = 'z';
         num_trimmed++;
      }
   }

   /* num_e = num_e - num_trimmed; */
   return(num_trimmed);
}

