/* setup.c contains routines to set the initial conditions */

#include "const.h"
#include "structs.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>

extern double start_time, end_time;
extern SPACE sp;
extern SPACE particle_trim;
extern SPACE element_trim;
extern int num_e, num_eg, num_p, num_pg, num_o;
extern ELEMENT e[];
extern ELEMENT_GENERATOR eg[];
extern particle_pointer particle_head;
extern PARTICLE_GENERATOR pg[];
extern OBJECT o[];

void add_freestream(double, double);

int set_up_space(double xsize,double ysize) {

   double ptrimw = 0.1;
   double etrimw = 1.0;

   /* if the size is not positive and significant */
   if ((xsize <= SM_THRESH) || (ysize <= SM_THRESH)) return(1);

   /* set origin and size */
   sp.o.x[0] = -xsize/2;
   sp.o.x[1] = -ysize/2;
   sp.s.x[0] = xsize;
   sp.s.x[1] = ysize;

   particle_trim.o.x[0] = sp.o.x[0] - ptrimw;
   particle_trim.o.x[1] = sp.o.x[1] - ptrimw;
   particle_trim.s.x[0] = sp.s.x[0] + 2*ptrimw;
   particle_trim.s.x[1] = sp.s.x[1] + 2*ptrimw;

   element_trim.o.x[0] = sp.o.x[0] - etrimw;
   element_trim.o.x[1] = sp.o.x[1] - etrimw;
   element_trim.s.x[0] = sp.s.x[0] + 2*etrimw;
   element_trim.s.x[1] = sp.s.x[1] + 2*etrimw;

   return(0);
}


/*
 * This routine takes an index as input and creates the
 * elements, particles, generators, etc. that make up a
 * certain initial problem definition. The index means:
 *  0  nothing
 *  1  two vortexes
 *  2  10 interacting vortexes
 *  3  a freestream and vortex generator
 *  4  smoke rising from a smouldering source
 */
int set_up_stuff(int index) {

   int i;
   particle_head = NULL;

   num_e = 0;	/* number of elements */
   num_eg = 0;	/* number of generators */
   num_p = 0;	/* number of particles */
   num_pg = 0;	/* number of particle generators */
   num_o = 0;	/* number of objects */

   if (index == 0) {		/* no objects at all */
   } else if (index == 1) {	/* a very simple exercise */
      num_e = 2;
      num_eg = 0;
      num_p = 0;
      num_o = 0;
      e[0].index = -1;
      e[0].loc.x[0] = -1.0;
      e[0].loc.x[1] = 0.0;
      e[0].type = 'v';
      e[0].str = 1.0;
      e[0].val = 0.0;
      e[1].index = -1;
      e[1].loc.x[0] = 1.0;
      e[1].loc.x[1] = 0.0;
      e[1].type = 'v';
      e[1].str = 1.0;
      e[1].val = 0.0;
      
   } else if (index == 2) {	/* some interacting vortexes */
      num_e = 10;
      num_eg = 0;
      num_p = 0;
      num_o = 0;
      for (i=0; i<num_e; i++) {
         e[i].index = -1;
         e[i].loc.x[0] = (sp.s.x[0]*rand())/RAND_MAX + sp.o.x[0];
         e[i].loc.x[1] = (sp.s.x[1]*rand())/RAND_MAX + sp.o.x[1];
         e[i].type = 'v';
         e[i].str = (0.5*rand())/RAND_MAX;
         e[i].val = 0.0;
      }

   } else if (index == 3) {	/* freestream and vortex generator */
      /* create these kinds of vortexes */
      eg[0].model.type = 'b';
      eg[0].model.str = -0.25;	/* units of strength per second */
      eg[0].model.val = 0.1;
      /* place generator near the center */
      eg[0].loc.x[0] = sp.o.x[0] + 1.0;
      eg[0].loc.x[1] = 0;
      eg[0].rate = 10;
      num_eg = 1;
      /* set the freestream, speed 1, to the right */
      (void) add_freestream(1.0,0.0);
      /* and create some particle emitters */
      pg[0].loc.x[0] = sp.o.x[0];
      pg[0].loc.x[1] = 0.2;
      pg[0].rad = 0.05;
      pg[0].rate = 140;
      pg[1].loc.x[0] = sp.o.x[0];
      pg[1].loc.x[1] = -0.2;
      pg[1].rad = 0.05;
      pg[1].rate = 100;
      num_pg = 2;
      for (i=1; i<30; i++) {
         e[num_e].index = 0;	/* they are fixed to ground */
         e[num_e].type = 'b';
         e[num_e].loc.x[0] = eg[0].loc.x[0] - 0.2*i;
         e[num_e].loc.x[1] = 0.0;
         e[num_e].str = -0.05;
         e[num_e].val = 0.1;
         num_e++;
      }

   } else if (index == 4) {	/* smoke rising from a smouldering source */
      (void) add_freestream(1.0,90.0);
      eg[0].model.type = 'b';
      eg[0].model.str = -0.1;	/* units of strength per second */
      eg[0].model.val = 0.01;
      eg[0].loc.x[0] = 0.1;
      eg[0].loc.x[1] = sp.o.x[1] + 0.2;
      eg[0].rate = 30;
      eg[1].model.type = 'b';
      eg[1].model.str = 0.1;	/* units of strength per second */
      eg[1].model.val = 0.01;
      eg[1].loc.x[0] = -0.1;
      eg[1].loc.x[1] = sp.o.x[1] + 0.2;
      eg[1].rate = 30;
      num_eg = 2;
      pg[0].loc.x[0] = 0.0;
      pg[0].loc.x[1] = sp.o.x[1];
      pg[0].rad = 0.04;
      pg[0].rate = 2000;
      pg[0].age_low = 1.0;
      pg[0].age_high = 2.0;
      num_pg = 1;
      for (i=1; i<20; i++) {
         e[num_e].index = 0;	/* they are fixed to ground */
         e[num_e].type = 'b';
         e[num_e].loc.x[0] = eg[0].loc.x[0];
         e[num_e].loc.x[1] = eg[0].loc.x[1] - 0.05*i;
         e[num_e].str = -0.005;
         e[num_e].val = 0.01;
         num_e++;
         e[num_e].index = 0;	/* they are fixed to ground */
         e[num_e].type = 'b';
         e[num_e].loc.x[0] = eg[1].loc.x[0];
         e[num_e].loc.x[1] = eg[1].loc.x[1] - 0.05*i;
         e[num_e].str = 0.005;
         e[num_e].val = 0.01;
         num_e++;
      }

   } else return (-1);

   return(0);
}

void add_random_vortexes(int num, double max_strength) {

   int i;

   for (i=num_e; i<num_e+num; i++) {
      e[i].index = -1;
      e[i].loc.x[0] = (sp.s.x[0]*rand())/RAND_MAX + sp.o.x[0];
      e[i].loc.x[1] = (sp.s.x[1]*rand())/RAND_MAX + sp.o.x[1];
      /* b is for vortex blob */
      e[i].type = 'b';
      e[i].str = (max_strength*rand())/RAND_MAX - max_strength/2.0;
      e[i].val = (2.5*rand())/RAND_MAX + 0.3;
   }

   num_e += num;
}


void add_banded_particles(int num, int out_size) {

   int i;
   particle_pointer new_particle;
   particle_pointer old_particle_head;
   double dw;		/* populated column width */
   register double xval;
   int col_width = 6;	/* width of columns (pixels) */

   dw = 2* sp.s.x[0] * col_width / out_size;

   for (i=0; i<num; i++) {
      new_particle = (PARTICLE *)malloc(sizeof(PARTICLE));
      xval = (sp.s.x[0]*rand())/RAND_MAX;
      /* reset the x-coord to create vertical banding */
      if ((int)(0.5+xval/dw) != (int)(xval/dw)) xval += dw/2;
      new_particle->loc.x[0] = xval + sp.o.x[0];
      new_particle->loc.x[1] = sp.s.x[1]*rand()/(RAND_MAX+1.0) + sp.o.x[1];
      new_particle->off_time = end_time + SM_THRESH; /* they never die */
      /* write a debug description of this particle */
      /* fprintf(stderr,"linked particle, loc= %lf %lf, off_time= %lf\n",new_particle->loc.x[0],new_particle->loc.x[1],new_particle->off_time); */
      /* add this particle to the head of the list! */
      old_particle_head = particle_head;
      particle_head = new_particle;
      new_particle->next_particle = old_particle_head;
      num_p++;
   }
}


void add_random_particles(int num) {

   int i;
   particle_pointer new_particle;
   particle_pointer old_particle_head;

   for (i=0; i<num; i++) {
      new_particle = (PARTICLE *)malloc(sizeof(PARTICLE));
      new_particle->loc.x[0] = sp.s.x[0]*rand()/(RAND_MAX+1.0) + sp.o.x[0];
      new_particle->loc.x[1] = sp.s.x[1]*rand()/(RAND_MAX+1.0) + sp.o.x[1];
      new_particle->off_time = end_time + SM_THRESH; /* they never die */
      /* write a debug description of this particle */
      /* fprintf(stderr,"linked particle, loc= %lf %lf, off_time= %lf\n",new_particle->loc.x[0],new_particle->loc.x[1],new_particle->off_time); */
      /* add this particle to the head of the list! */
      old_particle_head = particle_head;
      particle_head = new_particle;
      new_particle->next_particle = old_particle_head;
      num_p++;
   }
}


/*
 * When a freestream is defined, you use its speed and
 * direction, but those are converted to x and y 
 * components of velocity before calculation begins
 * The angle is to be defined in _degrees_, with zero
 * degrees being horizontally to the right
 */
void add_freestream(double speed, double angle) {

   e[num_e].index = 0;	/* attached to ground */
   e[num_e].type = 'f';
   e[num_e].str = speed;
   e[num_e].val = angle*PI/180.0;
   e[num_e].loc.x[0] = cos(e[num_e].val) * e[num_e].str;
   e[num_e].loc.x[1] = sin(e[num_e].val) * e[num_e].str;
   num_e ++;

}
