/* Write out the scene to text or Radiance or PGM file */

#include <stdio.h>
#include <stdlib.h>
#include "const.h"
#include "structs.h"

extern double curr_time;
extern SPACE sp;
extern int num_e, num_eg, num_p, num_pg, num_o;
extern ELEMENT e[];
extern particle_pointer particle_head;
extern PARTICLE_GENERATOR pg[];
extern OBJECT o[];


/*
 * This function writes basic usage information to stderr,
 * and then quits. Too bad.
 */
void usage(char progname[80],int status) {

   static char **cpp, *help_message[] =
   {
       "where options include:",
       "  -help          writes this help information",
       "  -text          writes output to stdout as text (default)",
       "  -radiance      writes output as Radiance file",
       "  -pgm           writes output as PGM image file",
       " ",
       "  -end val       sets simulation end time",
       "  -cdt val       sets time interval for each compute iteration",
       "  -odt val       sets time interval between output steps",
       " ",
       "  -size val val  sets solution dimensions, width and height",
       "  -width         sets the output width (pixels/columns),",
       "                 output height is inferred from size",
       " ",
       "  -nv val        adds val vortexes randomly over the area",
       "  -np val        adds val particles randomly over the area",
       "  -kv val        set kinematic viscosity, default is 1.46e-5",
       " ",
       "  -seed val      set the seed for the psuedo-random number generator",
       " ",
       "Options may be abbreviated to an unambiguous length.",
       NULL
   };

   fprintf(stderr, "usage:\n  %s [-options ...]\n\n", progname);
   for (cpp = help_message; *cpp; cpp++)
      fprintf(stderr, "%s\n", *cpp);
   exit(status);
}


/*
 * Writes a Radiance-readable .rad file, within are descriptions
 * of colors used, a groundplane at z=0, and all of the particles
 * and objects in the scene. Particles are spheres, objects are
 * joined cylinders or closed polygons.
 */
/* This routine is untested! */
int write_radiance(char filename[80]) {

   int i;
   FILE *fp;
   double time = curr_time;
   particle_pointer particle = particle_head;

   /* open the .rad file for writing */
   fp = fopen(filename,"w");

   /* write a header */
   fprintf(fp,"# Radiance description of 2D particle flow at time= %lf\n",time);
   fprintf(fp,"\n");
   fprintf(fp,"void plastic space_color 0 0 5  0.2 0.2 0.2  0.0 0.0\n");
   fprintf(fp,"void plastic particle_color 0 0 5  0.5 0.5 0.5  0.0 0.0\n");
   fprintf(fp,"\n");

   /* write a description of the space */
   fprintf(fp,"space_color polygon the_space 0 0 12\n");
   fprintf(fp,"%lf %lf %lf\n",sp.o.x[0], sp.o.x[1], 0.0);
   fprintf(fp,"%lf %lf %lf\n",sp.o.x[0]+sp.s.x[0], sp.o.x[1], 0.0);
   fprintf(fp,"%lf %lf %lf\n",sp.o.x[0]+sp.s.x[0], sp.o.x[1]+sp.s.x[1], 0.0);
   fprintf(fp,"%lf %lf %lf\n",sp.o.x[0], sp.o.x[1]+sp.s.x[1], 0.0);
   fprintf(fp,"\n");

   /* write a description of the particles */
   while (particle->next_particle != NULL) {
      if (particle->off_time > curr_time) {
         fprintf(fp,"particle_color sphere p%d\n",i);
         fprintf(fp,"0 0 4  %lf %lf %lf  %lf\n",particle->loc.x[0],particle->loc.x[1], 0.1, 0.025);
         fprintf(fp,"\n");
      }
      particle = particle->next_particle;
   }

   /* write a description of the objects */
   for (i=0; i<num_o; i++) {
      fprintf(fp,"# object %d\n",i);
   }

   /* close the file before returning to main */
   (void) fclose(fp);

   /* everything went OK, return zero */
   return(0);
}


/*
 * Write a Poskanzer GrayMap (.PGM) file using the locations
 * of all of the particles. I'd imagine this looks best with 
 * tons and tons of particles, but I'm setting the maxval to 
 * 8, so that nothing is too dark, and the density map stays
 * constant.
 */
int write_pgm(char filename[80],int width) {

   int i,j;
   FILE *fp;
   char onerow[MAX_OUTPUT_SIZE][MAX_OUTPUT_SIZE];
   register char raw_val;
   register double ox, oy, sx, sy;
   int xloc, yloc;
   /* char maxval = 4; */
   int height;			/* number of rows high */
   int scale;			/* scaling factor */
   double time = curr_time;
   particle_pointer particle = particle_head;

   /* limit the number of columns */
   if (width > MAX_OUTPUT_SIZE) width = MAX_OUTPUT_SIZE;

   /* compute the number of rows necessary */
   height = (sp.s.x[1]/sp.s.x[0])*width;
   if (height > MAX_OUTPUT_SIZE) {
      width = (width*MAX_OUTPUT_SIZE)/height;
      height = MAX_OUTPUT_SIZE;
   }

   /* calculate the pixel intensity scaling factor based on
    * the number of particles and the size of the image */
   /* if (num_p < 1000) scale = 100;
   else scale = 60*width*height/num_p; */
   scale = 10;

   /* initialize the character arrays */
   for (i=0; i<height; i++) {
      for (j=0;j<width;j++) onerow[i][j] = 0;
   }

   /* set some useful constants */
   ox = sp.o.x[0];
   oy = sp.o.x[1];
   sx = sp.s.x[0];
   sy = sp.s.x[1];

   /* increment a pixel for each particle in the linked list */
   if (particle != NULL) while (particle->next_particle != NULL) {
      if (particle->off_time > curr_time) {
         /* find out where that particle lands */
         xloc = ((particle->loc.x[0] - ox) / sx) * width;
         yloc = ((particle->loc.x[1] - oy) / sy) * height;
         if (xloc < width && xloc > -1 && yloc < height && yloc > -1) {
            onerow[yloc][xloc]++;
         }
      }
      particle = particle->next_particle;
   }

   /* open the .pgm file for writing */
   fp = fopen(filename,"w");

   /* write a header */
   fprintf(fp,"P2\n# 2D particle flow at time= %lf\n",time);
   fprintf(fp,"%d %d\n%d\n",width, height, 255);

   /* write each data point as an int to the file */
   for (i=height-1; i>-1; i--) {
      for (j=0; j<width; j++) {
         raw_val = onerow[i][j];
         /* write the intensity based on the number of particles
          * in that cell, but never exceed 255, the limit of a
          * binary PGM, AND the maxval necessary to use these
          * files as input to mpeg_encode */
         if (raw_val*scale > 255) fprintf(fp,"255\n");
         else fprintf(fp,"%d\n",onerow[i][j]*scale);
      }
   }

   /* close the file before returning to main */
   (void) fclose(fp);

   /* everything went OK, return zero */
   return(0);
}


/*
 * Draw a textual representation of the particles and objects
 * for each frame and send it to stdout. The height of the text
 * array is determined by proportionally scaling the solution 
 * space to the width in columns
 *
 * if what_to_write is zero, write everything; if 1, write
 * only particles; if 2, write only elements
 */
int write_text(int what_to_write,int columns) {

   int rows;
   int i,j;
   char onerow[100][101];
   register double ox, oy, sx, sy;
   int xloc, yloc;
   double time = curr_time;
   particle_pointer particle = particle_head;

   /* limit the number of columns */
   if (columns > 100) columns = 100;

   /* compute the number of rows necessary */
   rows = (sp.s.x[1]/sp.s.x[0])*columns;
   if (rows > 100) {
      columns = (columns*100)/rows;
      rows = 100;
   }

   /* initialize the character arrays */
   for (i=0; i<rows; i++) {
      for (j=0;j<columns;j++) onerow[i][j] = '.';
      onerow[i][columns] = '\0';
   }

   /* set some useful constants */
   ox = sp.o.x[0];
   oy = sp.o.x[1];
   sx = sp.s.x[0];
   sy = sp.s.x[1];

   /* darken the respective character if the element resides
    * in its area */
   if (what_to_write == 0 || what_to_write == 2) {
      for (i=0; i<num_e; i++) if (e[i].type != 'f') {
         /* find out where that element lands */
         xloc = ((e[i].loc.x[0] - ox) / sx) * columns;
         yloc = ((e[i].loc.x[1] - oy) / sy) * rows;
         if (xloc < columns && xloc > -1 && yloc < rows && yloc > -1) {
            if (onerow[yloc][xloc] == '.') onerow[yloc][xloc] = '1';
            else if (onerow[yloc][xloc] >= '0') onerow[yloc][xloc]++;
            else if (onerow[yloc][xloc] == '9') onerow[yloc][xloc] = 'A';
            else onerow[yloc][xloc]++;
         }
      }
   }

   /* write a char for each particle in the linked list */
   if ((particle != NULL) && (what_to_write == 0 || what_to_write == 1)) {
      while (particle->next_particle != NULL) {
         if (particle->off_time > curr_time) {
            /* find out where that particle lands */
            xloc = ((particle->loc.x[0] - ox) / sx) * columns;
            yloc = ((particle->loc.x[1] - oy) / sy) * rows;
            if (xloc < columns && xloc > -1 && yloc < rows && yloc > -1) {
               if (onerow[yloc][xloc] == '.') onerow[yloc][xloc] = '1';
               else if (onerow[yloc][xloc] < '9') onerow[yloc][xloc]++;
               /* else if (onerow[yloc][xloc] == '9') onerow[yloc][xloc] = 'A'; */
               /* else onerow[yloc][xloc]++; */
            }
         }
         particle = particle->next_particle;
      }
   }

   /* write all the characters to stdout */
   fprintf(stdout,"\nOutput from vortex2d at time= %lf\n",time);
   for (i=rows-1; i>-1; i--) fprintf(stdout,"%s\n",onerow[i]);
   /* fflush(stdout); */
}


void write_error(char message[80],int code) {

   fprintf(stderr,"%s\n",message);
   exit(code);
}


void write_finish() {

   fprintf(stderr,"\nFinished simulation.\n");
   exit(0);
}


void write_element_list() {

   int i;
   for (i=0; i<num_e; i++) {
      fprintf(stderr,"Element type %c, location %lf %lf, strength %lf\n",e[i].type, e[i].loc.x[0], e[i].loc.x[1], e[i].str);
   }
}


void write_particle_list() {

   int count = 0;
   particle_pointer particle = particle_head;

   while (particle->next_particle != NULL) {
      fprintf(stderr,"Particle %d, location %lf %lf, age %lf\n",count,particle->loc.x[0],particle->loc.x[1],particle->off_time);
      particle = particle->next_particle;
      count++;
   }
}


void write_emitter_list() {

   int i;
   for (i=0; i<num_pg; i++) {
      fprintf(stderr,"Particle generator %d, location %lf %lf, rate %lf\n",i,pg[i].loc.x[0],pg[i].loc.x[1],pg[i].rate);
   }
}


void write_stats() {

   fprintf(stderr,"num_e= %d, num_eg= %d, num_p= %d, num_pg= %d, num_o= %d\n",num_e,num_eg,num_p,num_pg,num_o);

}
